@@ -71,6 +71,17 @@ static QString page2string(fz_text_page *page){
 	return str;
 }
 
+static QString block2string(fz_text_block *block, bool addNewLines){
+	fz_text_line *line;
+	fz_text_span *span;
+	QString str("");
+	for (line = block->lines; line < block->lines + block->len; line++){
+		for (span = line->spans; span < line->spans + line->len; span++)
+			str += textChar_p2string(span->text,span->len) + (addNewLines?"\n":"");// pseudo-newline
+	}
+	return str;
+}
+
 static QList<QRect> searchInPage(fz_text_page *page,QString search,bool caseSensitive){
 	fz_text_block *block;
 	fz_text_line *line;
@@ -122,6 +133,8 @@ static QList<QRect> textAreas(fz_text_page *page,int zoom){
 	QRect nextR;
 	for (block = page->blocks; block < page->blocks + page->len; block++){
 		nextR = fz_rectToQRect(block->bbox);
+		if(block2string(block,false).contains(QRegExp("^\\s*[0-9ivxlcdmIVXLCDM]*\\s*$")))
+			continue;//page number?
 		if(commonWidth(uni,nextR) < 0.3*std::min(uni.width(),nextR.width())){//new column
 			textAreas += uni;
 			uni = nextR;
@@ -182,26 +195,35 @@ QVariantList outlineToVariantList(fz_outline* outline){
 	return outlList;
 }
 
+QRectF pdfRectToRectF(pdf_obj* pdf_rect){
+	return QRectF (pdf_to_real(pdf_array_get(pdf_rect,0)),
+		pdf_to_real(pdf_array_get(pdf_rect,1)),
+		pdf_to_real(pdf_array_get(pdf_rect,2))-
+			pdf_to_real(pdf_array_get(pdf_rect,0)),
+		pdf_to_real(pdf_array_get(pdf_rect,3))-
+			pdf_to_real(pdf_array_get(pdf_rect,01))
+	);
+}
 
+//uses PDF coordinate system, i.e. up is positive
+//don't use bottomLeft etc., they are misleading!
 pdf_obj* rectFToPdfRect(QRectF rect,fz_context* context){
 	pdf_obj* ret = pdf_new_array(context, 4/*int initialcap*/);
-	pdf_array_push(ret, pdf_new_real(context,rect.bottomLeft().x()));
-	pdf_array_push(ret, pdf_new_real(context,rect.bottomLeft().y()));
-	pdf_array_push(ret, pdf_new_real(context,rect.topRight().x()));
-	pdf_array_push(ret, pdf_new_real(context,rect.topRight().y()));
-	
+	pdf_array_push(ret, pdf_new_real(context,rect.x()));
+	pdf_array_push(ret, pdf_new_real(context,rect.y()));
+	pdf_array_push(ret, pdf_new_real(context,rect.x()+rect.width()));
+	pdf_array_push(ret, pdf_new_real(context,rect.y()+rect.height()));
 	
 	return ret;
 }
 
 
+PdfAnnotation::PdfAnnotation(QString st): subtype(st), state(NEW), originalPdfObj(NULL) {}
 
-PdfInkAnnotation::PdfInkAnnotation() : state(NEW), originalPdfObj(NULL) {}
 
-#pragma GCC diagnostic ignored "-Wwrite-strings"
-PdfInkAnnotation::PdfInkAnnotation(pdf_obj *anno) : state(UNCHANGED)  {
+PdfAnnotation::PdfAnnotation(pdf_obj *anno):  state(UNCHANGED)  {
 	originalPdfObj = anno;
-	pdf_obj* acol = pdf_dict_gets(anno,"C");
+	pdf_obj* acol = pdf_dict_gets(anno,const_cast<char*>("C"));
 	if(acol && pdf_array_len(acol) == 3){//RGB
 		col.setRgbF(pdf_to_real(pdf_array_get(acol,0)),
 			pdf_to_real(pdf_array_get(acol,1)),
@@ -213,9 +235,91 @@ PdfInkAnnotation::PdfInkAnnotation(pdf_obj *anno) : state(UNCHANGED)  {
 			pdf_to_real(pdf_array_get(acol,3)));
 	} else if (acol && pdf_array_len(acol) == 1)//Gray
 		col.setHslF(0,0,pdf_to_real(pdf_array_get(acol,0)));
-	else col.setAlpha(0);//or col.setHsl(0,0,0)? transparent
+	else
+		col.setAlpha(0);//or col.setHsl(0,0,0)? transparent
+	contents = pdf_to_str_buf(pdf_dict_gets(anno,const_cast<char*>("Contents")));
+	rect_prv = pdfRectToRectF(pdf_dict_gets(anno,const_cast<char*>("Rect")));
+}
 
-	pdf_obj* subpathsArr = pdf_dict_gets(anno,"InkList");
+PdfAnnotation::PdfAnnotation(const PdfAnnotation &copy) {
+	subtype = copy.subtype;
+	rect_prv = copy.rect_prv;
+	date = copy.date;
+	col = copy.col;
+	contents = copy.contents;
+	originalPdfObj = copy.originalPdfObj;
+	state = copy.state;
+}
+
+PdfAnnotation& PdfAnnotation::operator=(PdfAnnotation &copy){
+	subtype = copy.subtype;
+	rect_prv = copy.rect_prv;
+	date = copy.date;
+	col = copy.col;
+	contents = copy.contents;
+	originalPdfObj = copy.originalPdfObj;
+	state = copy.state;
+	return *this;
+}
+
+
+pdf_obj* PdfAnnotation::toPdfObj(fz_context *context){
+	if(state == UNCHANGED)
+		return originalPdfObj;
+	pdf_obj* ret = originalPdfObj ? originalPdfObj : pdf_new_dict(context, 5/*int initialcap*/);
+	pdf_dict_puts(ret,const_cast<char*>("Subtype"), pdf_new_name(context,subtype.toUtf8().data()));
+	pdf_dict_puts(ret,const_cast<char*>("Contents"), pdf_new_string(context, contents.toUtf8().data(), contents.toUtf8().length()));
+	
+	pdf_dict_puts(ret, const_cast<char*>("Rect"), rectFToPdfRect(rect(),context));
+	if(date.isValid()){
+		QString datestring = "D:" + date.toString("yyyyMMddhhmmss");
+// 		QDateTime utc = date.toUTC();
+// 		utc.setTimeSpec(Qt::LocalTime);//make them comparable
+// 		int hoffset = utc.secsTo(date) / 3600;
+// 		int moffset = (utc.secsTo(date) % 3600) /60;
+//  		datestring += (hoffset+moffset > 0?"+":"") + QString("%1'%2").arg(hoffset).arg(moffset);
+		pdf_dict_puts(ret,const_cast<char*>("M"), pdf_new_string(context, datestring.toUtf8().data(), datestring.toUtf8().length()));
+	}
+	
+	if(col.alphaF() == 0)
+		return ret;
+	pdf_obj* color = pdf_new_array(context, 3/*RGB*/);
+	pdf_array_push(color, pdf_new_real(context,col.redF()));
+	pdf_array_push(color, pdf_new_real(context,col.greenF()));
+	pdf_array_push(color, pdf_new_real(context,col.blueF()));
+	pdf_dict_puts(ret,const_cast<char*>("C"),color);
+	
+	return ret;
+}
+
+
+PdfAnnotation* PdfAnnotation::fromPdfObj(pdf_obj *anno){
+	QString subtype = pdf_to_name(pdf_dict_gets(anno,const_cast<char*>("Subtype")));
+	if(subtype == "Ink")
+		return new PdfInkAnnotation(anno);
+	else if (subtype == "FreeText")
+		return new PdfFreeTextAnnotation(anno);
+	else if (subtype == "Text")
+		return new PdfTextAnnotation(anno);
+	else 
+		return new PdfAnnotation(anno);
+}
+
+QRectF PdfAnnotation::rect(){
+	return rect_prv;
+}
+
+void PdfAnnotation::setRect(QRectF r){
+	rect_prv = r;
+}
+
+PdfInkAnnotation::PdfInkAnnotation() : PdfAnnotation("Ink") {}
+
+
+PdfInkAnnotation::PdfInkAnnotation(pdf_obj *anno) : PdfAnnotation(anno)  {
+	subtype = "Ink";
+	
+	pdf_obj* subpathsArr = pdf_dict_gets(anno,const_cast<char*>("InkList"));
 	for(int i=0;i<pdf_array_len(subpathsArr);i++){
 		pdf_obj* subpath = pdf_array_get(subpathsArr,i);
 		if(pdf_array_len(subpath) >= 2)
@@ -226,45 +330,22 @@ PdfInkAnnotation::PdfInkAnnotation(pdf_obj *anno) : state(UNCHANGED)  {
 				pdf_to_real(pdf_array_get(subpath,j+1)));
 	}
 }
-#pragma GCC diagnostic warning "-Wwrite-strings"
 
-PdfInkAnnotation::PdfInkAnnotation(const PdfInkAnnotation &copy){
+
+PdfInkAnnotation::PdfInkAnnotation(const PdfInkAnnotation &copy) : PdfAnnotation(copy) {
 	path = copy.path;
-	date = copy.date;
-	col = copy.col;
-	contents = copy.contents;
-	originalPdfObj = copy.originalPdfObj;
-	state = copy.state;
 }
 
 
 PdfInkAnnotation& PdfInkAnnotation::operator=(PdfInkAnnotation &copy){
+	PdfAnnotation::operator=(copy);
 	path = copy.path;
-	date = copy.date;
-	col = copy.col;
-	contents = copy.contents;
-	originalPdfObj = copy.originalPdfObj;
-	state = copy.state;
 	return *this;
 }
 
-#pragma GCC diagnostic ignored "-Wwrite-strings"
 //see iTexts: PdfAnnotation::createInk
 pdf_obj* PdfInkAnnotation::toPdfObj(fz_context *context){
-	if(state == UNCHANGED)
-		return originalPdfObj;
-	pdf_obj* ret = originalPdfObj ? originalPdfObj : pdf_new_dict(context, 5/*int initialcap*/);
-	pdf_dict_puts(ret, "Rect", rectFToPdfRect(path.boundingRect(),context));
-	pdf_dict_puts(ret,"Subtype", pdf_new_name(context,"Ink"));
-	pdf_dict_puts(ret,"Contents", pdf_new_string(context, contents.toUtf8().data(), contents.toUtf8().length()));
-	
-	
-	pdf_obj* color = pdf_new_array(context, 3/*RGB*/);
-	pdf_array_push(color, pdf_new_real(context,col.redF()));
-	pdf_array_push(color, pdf_new_real(context,col.greenF()));
-	pdf_array_push(color, pdf_new_real(context,col.blueF()));
-	pdf_dict_puts(ret,"C",color);
-	
+	pdf_obj* ret = PdfAnnotation::toPdfObj(context);
 	
 	QList<QPolygonF> subpaths = path.toSubpathPolygons();
 	pdf_obj* subpathsArr = pdf_new_array(context, subpaths.length()/*int initialcap*/);
@@ -276,10 +357,79 @@ pdf_obj* PdfInkAnnotation::toPdfObj(fz_context *context){
 		}
 		pdf_array_push(subpathsArr,subpath);
 	}
-	pdf_dict_puts(ret,"InkList",subpathsArr);
+	pdf_dict_puts(ret,const_cast<char*>("InkList"),subpathsArr);
+	return ret;
+}
+
+
+QRectF PdfInkAnnotation::rect(){
+	return path.boundingRect();
+}
+
+PdfFreeTextAnnotation::PdfFreeTextAnnotation() : PdfAnnotation("FreeText"), defApp("/Helvetica 16 Tf\n1 0 0 rg"), defStyle("font: Helvetica,sans-serif 16.0pt; text-align:left; color:#FF0000") {
+	col = QColor("transparent");
+	rect_prv = QRectF(10,10,30,30);
+	contents = "Test";
+}
+
+
+PdfFreeTextAnnotation::PdfFreeTextAnnotation(pdf_obj *anno) : PdfAnnotation(anno)  {
+	subtype = "FreeText";
+	defApp = pdf_to_str_buf(pdf_dict_gets(anno,const_cast<char*>("DA")));
+	defStyle = pdf_to_str_buf(pdf_dict_gets(anno,const_cast<char*>("DS")));
+}
+
+
+PdfFreeTextAnnotation::PdfFreeTextAnnotation(const PdfFreeTextAnnotation &copy) : PdfAnnotation(copy) {
+	defApp = copy.defApp;
+	defStyle = copy.defStyle;
+}
+
+
+PdfFreeTextAnnotation& PdfFreeTextAnnotation::operator=(PdfFreeTextAnnotation &copy){
+	PdfAnnotation::operator=(copy);
+	defApp = copy.defApp;
+	defStyle = copy.defStyle;
+	return *this;
+}
+
+
+pdf_obj* PdfFreeTextAnnotation::toPdfObj(fz_context *context){
+	pdf_obj* ret = PdfAnnotation::toPdfObj(context);
+	
+	pdf_dict_puts(ret,const_cast<char*>("DA"),pdf_new_string(context,defApp.toUtf8().data(),defApp.length()));
+	pdf_dict_puts(ret,const_cast<char*>("DS"),pdf_new_string(context,defStyle.toUtf8().data(),defStyle.length()));
+	return ret;
+}
+
+
+PdfTextAnnotation::PdfTextAnnotation() : PdfAnnotation("Text"), open(true) {
+}
+
+
+PdfTextAnnotation::PdfTextAnnotation(pdf_obj *anno) : PdfAnnotation(anno)  {
+	subtype = "Text";
+	open = pdf_to_bool(pdf_dict_gets(anno,const_cast<char*>("Open")));
+}
+
+
+PdfTextAnnotation::PdfTextAnnotation(const PdfTextAnnotation &copy) : PdfAnnotation(copy) {
+	open = copy.open;
+}
+
+
+PdfTextAnnotation& PdfTextAnnotation::operator=(PdfTextAnnotation &copy){
+	PdfAnnotation::operator=(copy);
+	open = copy.open;
+	return *this;
+}
+
+
+pdf_obj* PdfTextAnnotation::toPdfObj(fz_context *context){
+	pdf_obj* ret = PdfAnnotation::toPdfObj(context);
+	pdf_dict_puts(ret,const_cast<char*>("Open"),pdf_new_bool(context,open));
 	return ret;
 }
-#pragma GCC diagnostic warning "-Wwrite-strings"
 
 PdfLink::PdfLink(QObject *parent) : QObject(parent) {
 	type_prv = 0;
@@ -336,6 +486,7 @@ QDeclarativeItem(parent) {
 	setAcceptedMouseButtons(Qt::LeftButton | Qt::RightButton | Qt::MidButton);
 	qRegisterMetaType<PdfLink>();
 	qmlRegisterType<PdfLink>();
+	curRegionIdx = -1;
 	context = NULL;
 	document = NULL;
 	pixmap = NULL;
@@ -357,6 +508,11 @@ QDeclarativeItem(parent) {
 	textDocument = QList<QString>();
 	hitCount = QList<int>();
 	
+	QVector<QRgb> ct;
+	ct << QColor(137,153,137,255).rgba ();
+	ct << 0;
+	noteImage = QImage(":/note_inv.png").scaled(24,24).convertToFormat(QImage::Format_Indexed8, ct);
+	
 	image = new QImage(800, 800, QImage::Format_RGB32);
 	QRgb value;
 	value = qRgb(189, 149, 39); // 0xffbd9527
@@ -379,7 +535,7 @@ QPdfImage::~QPdfImage(){
 	}
 }
 
-#pragma GCC diagnostic ignored "-Wwrite-strings"
+
 bool QPdfImage::loadFile(QString filename){
 	if(!QFile::exists(filename))
 		return false;
@@ -395,6 +551,9 @@ bool QPdfImage::loadFile(QString filename){
 		fz_free_text_page(context,text_page);
 		text_page = NULL;
 	}
+	textDocument.clear();
+	regions.clear();
+	curRegionIdx = -1;
 	//NOTE:filename_prv has to be the old one:
 	writeAnnotationsToPage();
 	pageobj = NULL;
@@ -411,22 +570,22 @@ bool QPdfImage::loadFile(QString filename){
 	
 	if(!context)
 		context = fz_new_context(NULL, NULL, FZ_STORE_UNLIMITED);
-	textDocument.clear();
+	
 	
 	pdf_document* xref = pdf_open_document(context, filename.toLocal8Bit().data());
 	document = (fz_document*) xref;//fz_document is first member of pdf_document
 
 
-	pdf_obj *info = pdf_dict_gets(xref->trailer,"Info");
+	pdf_obj *info = pdf_dict_gets(xref->trailer,const_cast<char*>("Info"));
 	
-	pdf_obj *titleObj = pdf_dict_gets(info,"Title");
+	pdf_obj *titleObj = pdf_dict_gets(info,const_cast<char*>("Title"));
 	if(titleObj)
 		title_prv = pdf_to_str_buf(titleObj);
 	else 
 		title_prv = "";
 	emit titleChanged(title_prv);
 	
-	pdf_obj *authorObj = pdf_dict_gets(info,"Author");
+	pdf_obj *authorObj = pdf_dict_gets(info,const_cast<char*>("Author"));
 	if(authorObj)
 		author_prv = pdf_to_str_buf(authorObj);
 	else 
@@ -441,41 +600,38 @@ bool QPdfImage::loadFile(QString filename){
 	emit outlineChanged();
 	return true;
 }
-#pragma GCC diagnostic warning "-Wwrite-strings"
 
-#pragma GCC diagnostic ignored "-Wwrite-strings"
 void QPdfImage::writeAnnotationsToPage(){
 	if(!pageobj || !pageModifications)
 		return;
 	fileModified = true;
 	pageModifications = 0;
 	// before: old annotsArr: A1 A2  A3  ... An
-	// now: inkAnnos:     A1 A2* A3- ... An An+1 An+2 ... Am
+	// now: annos:     A1 A2* A3- ... An An+1 An+2 ... Am
 	//*modified
 	//-deleted (only tag, still in list!)
 	//n:length of annotsArr
-	//m:length of inkAnnos (m>=n)
+	//m:length of annos (m>=n)
 	
 	pdf_obj *annotsArr = pdf_new_array(context, 1);
 	
-	for(int i=0;i < inkAnnos.length();i++)
-		if(inkAnnos[i].state != PdfInkAnnotation::DELETED)
-			pdf_array_push(annotsArr,inkAnnos[i].toPdfObj(context));
+	for(int i=0;i < annos.length();i++)
+		if(annos[i]->state != PdfAnnotation::DELETED)
+			pdf_array_push(annotsArr,annos[i]->toPdfObj(context));
 
-	pdf_dict_puts(pageobj, "Annots", annotsArr);
+	pdf_dict_puts(pageobj, const_cast<char*>("Annots"), annotsArr);
 }
 
 void QPdfImage::readAnnotationsFromPage(){
-	inkAnnos.clear();
-	pdf_obj *annotsArr = pdf_dict_gets(pageobj, "Annots");
+	annos.clear();
+	pdf_obj *annotsArr = pdf_dict_gets(pageobj, const_cast<char*>("Annots"));
 	if(annotsArr  && pdf_is_array(annotsArr)){
 		for(int i=0;i<pdf_array_len(annotsArr);i++){
-			if(pdf_to_name(pdf_dict_gets(pdf_array_get(annotsArr,i),"Subtype")) == QString("Ink"))
-				inkAnnos << PdfInkAnnotation(pdf_array_get(annotsArr,i));
+			//if(pdf_to_name(pdf_dict_gets(pdf_array_get(annotsArr,i),"Subtype")) == QString("Ink"))
+				annos << PdfAnnotation::fromPdfObj(pdf_array_get(annotsArr,i));
 		}
 	}
 }
-#pragma GCC diagnostic warning "-Wwrite-strings"
 
 
 void QPdfImage::writeCurrentFile(){
@@ -544,6 +700,11 @@ bool QPdfImage::gotoPage(int pagenum,bool render){
 	highlight += searchInPage(text_page,highlightPhrase_prv,caseSensitive_prv);
 	regions.clear();
 	regions += textAreas(text_page,zoom_prv);
+	if(regions.length() > 0){
+		if(curRegionIdx == -1)
+			curRegionIdx = 0;
+	} else
+		curRegionIdx = -1;
 	
 	fz_link* fz_links = fz_load_links(document, page);
 	links_prv.clear();
@@ -616,6 +777,17 @@ void QPdfImage::paint(QPainter *painter, const QStyleOptionGraphicsItem*, QWidge
 		ql->render(painter);
 		return;
 	}
+	if(grainyMode_prv){
+// 		painter->setPen(QColor ( 0, 0, 0, 255));
+// 		painter->setBrush(QBrush(QColor ("transparent")));
+// 		for (fz_text_block *block = text_page->blocks; block < text_page->blocks + text_page->len; block++){
+// 			painter->drawRect(fz_rectToQRect(block->bbox));
+// 		}
+		for(int i=0;i < regions.length();i++){
+			painter->drawRect(regions[i]);
+		}
+		return;
+	}
 	if(image)
 		painter->drawImage (QPoint(0,0),*image);
 	painter->setPen(QColor("transparent"));
@@ -637,13 +809,46 @@ void QPdfImage::paint(QPainter *painter, const QStyleOptionGraphicsItem*, QWidge
 	
 	
 	painter->setBrush(QBrush(QColor ("transparent")));
-	painter->translate(0,height());
-	painter->scale(zoom_prv/100.0,-zoom_y_prv/100.0);
-	for(int i=0;i<inkAnnos.length();i++){
-		if(inkAnnos[i].state == PdfInkAnnotation::DELETED)
+	for(int i=0;i<annos.length();i++){
+		if(annos[i]->state == PdfAnnotation::DELETED)
 			continue;
-		painter->setPen(inkAnnos[i].col);
-		painter->drawPath(inkAnnos[i].path);
+		painter->setPen(annos[i]->col);
+		if(annos[i]->subtype == "Ink"){
+			painter->save();
+			painter->translate(0,height());
+			painter->scale(zoom_prv/100.0,-zoom_y_prv/100.0);
+			painter->drawPath(((PdfInkAnnotation*) annos[i])->path);
+			painter->restore();
+		} else if(annos[i]->subtype == "FreeText"){
+			PdfFreeTextAnnotation* ftAnno  = (PdfFreeTextAnnotation*) annos[i];
+			QRectF r = ftAnno->rect();
+			painter->drawText(QRectF(
+					r.x()*zoom_prv/100.0,
+					height() - (r.y()+r.height())*zoom_y_prv/100.0,
+					r.width()*zoom_prv/100.0,
+					r.height()*zoom_y_prv/100.0),
+				ftAnno->contents);
+			int d = ftAnno->defApp.lastIndexOf("Tf");
+			QFont fnt = painter->font();
+			fnt.setPointSize(ftAnno->defApp.mid(d-3,2).toInt() >= 0 ? ftAnno->defApp.mid(d-3,2).toInt() : 12);
+			painter->setFont(fnt);
+		} else if(annos[i]->subtype == "Text"){
+			PdfTextAnnotation* noteAnno  = (PdfTextAnnotation*) annos[i];
+			QRectF r = noteAnno->rect();
+			noteImage.setColor(0,noteAnno->col.rgba());
+			painter->drawImage(QPointF(
+					r.x()*zoom_prv/100.0,
+					height() - (r.y()+r.height())*zoom_y_prv/100.0),noteImage);
+			if(!noteAnno->open)
+				continue;
+			painter->drawText(QRectF(
+					r.x()*zoom_prv/100.0+24,
+					height() - (r.y()+r.height())*zoom_y_prv/100.0+24,
+					100*zoom_prv/100.0,
+					100*zoom_y_prv/100.0),
+					Qt::TextWordWrap | Qt::TextDontClip,
+				noteAnno->contents);
+		}
 	}
 
 	
@@ -658,7 +863,8 @@ void QPdfImage::paint(QPainter *painter, const QStyleOptionGraphicsItem*, QWidge
 
 void QPdfImage::mousePressEvent(QGraphicsSceneMouseEvent *event){
 	if(drawMode_prv){
-		inkAnnos.last().path.moveTo(event->pos().x()*100/zoom_prv,originalHeight_prv - event->pos().y()*100/zoom_y_prv);
+		PdfInkAnnotation* inkAnno = (PdfInkAnnotation*)  annos.last();
+		inkAnno->path.moveTo(event->pos().x()*100/zoom_prv,originalHeight_prv - event->pos().y()*100/zoom_y_prv);
 		update();
 		return;
 	}
@@ -668,45 +874,69 @@ void QPdfImage::mousePressEvent(QGraphicsSceneMouseEvent *event){
 			originalHeight_prv - event->pos().y()*100/zoom_y_prv,
 			32*100/zoom_prv,
 			32*100/zoom_y_prv);
-		for(int i=0;i<inkAnnos.length();i++)
-			if(inkAnnos[i].path.intersects(clickRect)){
-				inkAnnos[i].state = PdfInkAnnotation::DELETED;
+		for(int i=0;i<annos.length();i++)
+			if(annos[i]->rect().intersects(clickRect)){
+				annos[i]->state = PdfInkAnnotation::DELETED;
 				pageModifications++;
 			}
 		update();
 		return;
 	}
+	if(noteMode_prv){
+		annos << new PdfTextAnnotation();
+		pageModifications++;//we'll add a free text annotation to the page
+		annos.last()->date = QDateTime::currentDateTime();
+		annos.last()->col = drawColor_prv;
+		annos.last()->setRect(QRectF(event->pos().x()*100/zoom_prv,
+			originalHeight_prv - event->pos().y()*100/zoom_y_prv,
+			20,
+			20));
+		return;
+	}
+	QPoint clickP = event->pos().toPoint();
 	for(int i=0;i<links_prv.length();i++){
-		if(links_prv[i].position().contains(event->pos().toPoint())){
+		if(links_prv[i].position().contains(clickP)){
 			emit linkSelected(links_prv[i].destination());
 // 			(qDebug() << "hit") << i << links_prv[i].destination().toInt(); 
 			return;
 		}
-// 			(qDebug() << "coords:") << i;
-// 			qDebug() << event->pos().x();
-// 			qDebug() << event->pos().y();
-// 			qDebug() << links_prv[i].position().x();
-// 			qDebug() << links_prv[i].position().y();
 	}
+	for(int i=0;i<annos.length();i++){
+		QRectF r = annos[i]->rect();
+
+		if(annos[i]->subtype == "Text" 
+			&& QRect(r.x()*zoom_prv/100.0,
+			height() - (r.y()+r.height())*zoom_y_prv/100.0,24,24).contains(clickP)){
+			
+			((PdfTextAnnotation*) annos[i])->open = !((PdfTextAnnotation*) annos[i])->open;
+			pageModifications++;
+			update();
+			return;
+		}
+	}
+	
 	
 	clickTimer.start();
 }
 
 void QPdfImage::mouseReleaseEvent(QGraphicsSceneMouseEvent*){
-	if(drawMode_prv || eraseMode_prv){
+	if(drawMode_prv || eraseMode_prv || noteMode_prv){
 		return;
 	}
 	clickTimer.stop();
 }
 
 void QPdfImage::mouseDoubleClickEvent(QGraphicsSceneMouseEvent*){
+	if(drawMode_prv || eraseMode_prv || noteMode_prv){
+		return;
+	}
 	clickTimer.stop();
 	emit clickForward();
 }
 
 void QPdfImage::mouseMoveEvent(QGraphicsSceneMouseEvent *event){
 	if(drawMode_prv){
-		inkAnnos.last().path.lineTo(event->pos().x()*100/zoom_prv,originalHeight_prv -event->pos().y()*100/zoom_y_prv);
+		((PdfInkAnnotation*) annos.last())->path.lineTo(event->pos().x()*100/zoom_prv,originalHeight_prv -event->pos().y()*100/zoom_y_prv);
 		update();
 		return;
 	}
@@ -774,7 +1004,7 @@ bool QPdfImage::searchNext(QString str){
 		if(i < regions.length())
 			curRegionIdx = i;
 		else
-			curRegionIdx = 0;
+			curRegionIdx = -1;
 	}
 	emit hitRectChanged();
 	return (bool) (hitOnPage+1);
@@ -815,7 +1045,7 @@ bool QPdfImage::searchPrev(QString str){
 		if(i < regions.length())
 			curRegionIdx = i;
 		else
-			curRegionIdx = 0;
+			curRegionIdx = -1;
 	}
 	emit hitRectChanged();
 	return (bool) (hitOnPage+1);
@@ -830,14 +1060,17 @@ QRect QPdfImage::prevRegion(bool render){
 		gotoPage(pagenum_prv-1,render);//sets regions
 	else
 		gotoPage(totalpages_prv,render);//sets regions
-	curRegionIdx = regions.length()-1;
+	curRegionIdx = regions.length()-1;//may be -1, i.e. invalid
 	
 	emit regionIdxChanged(curRegionIdx);
-	return regions[curRegionIdx];
+	return curRegion();
 }
 
 QRect QPdfImage::curRegion(){
-	return regions[curRegionIdx];
+	if(curRegionIdx >= 0)
+		return regions[curRegionIdx];
+	else //if no text is present, show the entire page
+		return QRect(0,0,width(),height());
 }
 
 QRect QPdfImage::nextRegion(bool render){
@@ -849,9 +1082,12 @@ QRect QPdfImage::nextRegion(bool render){
 		gotoPage(pagenum_prv+1,render);//sets regions
 	else
 		gotoPage(1,render);//sets regions
-	curRegionIdx = 0;
-	emit regionIdxChanged(0);
-	return regions[0];
+	if(regions.length() > 0)
+		curRegionIdx = 0;
+	else
+		curRegionIdx = -1;
+	emit regionIdxChanged(curRegionIdx);
+	return curRegion();
 }
 
 int QPdfImage::regionIdx(){
@@ -859,12 +1095,14 @@ int QPdfImage::regionIdx(){
 }
 
 QRect QPdfImage::setRegionIdx(int idx){
-	if(idx < regions.length())
+	if(idx < regions.length() && idx >= -1)
 		curRegionIdx = idx;
-	else
+	else if(regions.length() > 0)
 		curRegionIdx = 0;
+	else
+		curRegionIdx = -1;
 	emit regionIdxChanged(curRegionIdx);
-	return regions[curRegionIdx];
+	return curRegion();
 }
 
 
@@ -888,6 +1126,17 @@ void QPdfImage::setTextMode(bool tm){
 	emit textModeChanged(tm);
 }
 
+bool QPdfImage::grainyMode(){
+	return grainyMode_prv;
+}
+void QPdfImage::setGrainyMode(bool gm){
+	if(grainyMode_prv == gm)
+		return;
+	grainyMode_prv = gm;
+	update();
+	emit grainyModeChanged(gm);
+}
+
 int QPdfImage::textModeFontSize(){
 	return textModeFontSize_prv;
 }
@@ -990,18 +1239,19 @@ void QPdfImage::setDrawMode(bool dm){
 	if(dm == drawMode_prv)
 		return;
 	if(dm){
-		if(eraseMode_prv)
-			setEraseMode(false);
+		setEraseMode(false);
+		setNoteMode(false);
 		
-		inkAnnos << PdfInkAnnotation();
+		annos << new PdfInkAnnotation();
 		pageModifications++;//we'll add a path to the page (s.b.)
-		inkAnnos.last().col = drawColor_prv;
+		annos.last()->col = drawColor_prv;
+		annos.last()->date = QDateTime::currentDateTime();
 	}
 	drawMode_prv = dm;
 	emit drawModeChanged(dm);
 	if(!dm){
-		if(!inkAnnos.last().path.elementCount()){
-			inkAnnos.removeLast();//no, we don't: it's empty
+		if(!((PdfInkAnnotation*) annos.last())->path.elementCount()){
+			annos.removeLast();//no, we don't: it's empty
 			pageModifications--;//undo false increasement (s.a.)
 		}
 	}
@@ -1016,8 +1266,25 @@ void QPdfImage::setEraseMode(bool em){
 		return;
 	eraseMode_prv = em;
 	emit eraseModeChanged(em);
-	if(em && drawMode_prv)
+	if(em){
 		setDrawMode(false);
+		setNoteMode(false);
+	}
+}
+
+bool QPdfImage::noteMode(){
+	return noteMode_prv;
+}
+
+void QPdfImage::setNoteMode(bool ftm){
+	if(ftm == noteMode_prv)
+		return;
+	noteMode_prv = ftm;
+	emit noteModeChanged(ftm);
+	if(ftm){
+		setDrawMode(false);
+		setEraseMode(false);
+	}
 }
 
 bool QPdfImage::modified(){
@@ -1035,6 +1302,15 @@ void QPdfImage::setDrawColor(QColor cl){
 	drawColor_prv = cl;
 	emit drawColorChanged();
 }
+
+void QPdfImage::updateAnnoText(QString atext){
+	if(noteMode_prv){
+		((PdfTextAnnotation*) annos.last())->contents = atext;
+		update();
+		return;
+	}
+}
+
 QString QPdfImage::author(){
 	return author_prv;
 }
