@@ -160,4 +160,12 @@ QString SetHist::platform(){
 #elif defined Q_OS_WIN32
 	return "win32";
 #endif
+}
+
+QString SetHist::displayType(){
+#ifdef EINK_DISPLAY
+	return "eink";
+#else
+	return "normal";
+#endif
 }
\ No newline at end of file
